/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "OMS_app.h"
#include <math.h>
#include "Rogus.h"
#include "sched.h"

OMS_basic_app::
OMS_basic_app( long app_sig ) : as( app_sig )
// app_sig should be the application's creator
{
}

//--------------------------------------------------------------------------------------------------

DWORD OMS_app::our_start_time = 0;
static UINT default_output_ioref = 0;
static UINT default_input_ioref = 0;

void
OMS_app::
init_timer()
{
	if (timeBeginPeriod(1) != TIMERR_NOERROR) // set timer resolution to 1 millisecond
		ew.w("Cannot set timer resolution to 1 millisecond");
	our_start_time = timeGetTime();
}

c_queue<MyQueueElement> OMS_app::cq( 1000, "" ); // changed name from "input queue" to empty string 11/12/95 PP
parser *OMS_app::lp = 0;

static int binner(int n)
{
	return (int)(pow(2,n));
}

typedef UtilArray<MyMidiIn> MyMidiInArray;
typedef UtilPtr<MyMidiInArray> MyMidiInArrayPtr;
typedef UtilArray<MyMidiOut> MyMidiOutArray;
typedef UtilPtr<MyMidiOutArray> MyMidiOutArrayPtr;

static MyMidiInArrayPtr get_input_nodes()
{
	UINT n = midiInGetNumDevs();
	MyMidiInArrayPtr a = new MyMidiInArray(n);
	for (UINT i = 0; i < n; i++)
	{
		(*a)[i].open = 0;
		if (midiInGetDevCaps(i, &((*a)[i]), sizeof(MIDIINCAPS)) != MMSYSERR_NOERROR)
		{
			ew.w("Unable to get info for input device %d", i);
			strcpy((*a)[i].szPname, "Unknown");
		}
	}
	return a;
}

static MyMidiOutArrayPtr get_output_nodes()
{
	UINT n = midiOutGetNumDevs();
	MyMidiOutArrayPtr a = new MyMidiOutArray(n);
	for (UINT i = 0; i < n; i++)
	{
		(*a)[i].open = 0;
		if (midiOutGetDevCaps(i, &((*a)[i]), sizeof(MIDIOUTCAPS)) != MMSYSERR_NOERROR)
		{
			ew.w("Unable to get info for output device %d", i);
			strcpy((*a)[i].szPname, "Unknown");
		}
	}
	return a;
}

OMS_app::
OMS_app( long app_sig ) : OMS_basic_app( app_sig ), queueStat(10, binner, "Input Queue Usage")
{
	init_timer();

	inlh = get_input_nodes();	
	onlh = get_output_nodes();
//	OMS_node_info<MyMidiIn>::print_nodes(inlh, "input");
//	OMS_node_info<MyMidiOut>::print_nodes(onlh, "output");

	OMS_node_info<MyMidiIn>::select_by_name( default_input, inlh, &default_input_ioref);
	OMS_node_info<MyMidiOut>::select_by_name( default_output, onlh, &default_output_ioref);
}
/*
1. It's not necessary to ever close this connection since OMSSignOut will do it anyway.
3. Though a call to ew.e() will cause an exception to be thrown, meaning that OMS_app's dtor will 
not be called since it was not successfully constructed, stack unwinding will cause OMS_basic_app's 
dtor to be called since by the time we reach the body of OMS_app's ctor, OMS_basic_app's ctor has 
successfully executed.  The upshot of all of this is that OMSSignOut() will be called even if an 
error occurs in OMS_app's ctor.  See Stroustrup's "resource acquisition is initialization" scheme 
on p. 308 of _The C++ Programming Language, 2nd ed._
*/

 OMS_app::
~OMS_app()
{
	closeAllInputs();
	closeAllOutputs();
	timeEndPeriod(1);
}


ioref
OMS_app::
open_input( const char* name )
{
	err_ind e;
	MMRESULT mm;
	UINT n = default_input_ioref;
	e = OMS_node_info<MyMidiIn>::select_by_name( name, inlh, &n);
	if (e)
	{
		ew.w( "couldn't select an input node" ); return 0;
	}
	if ((*inlh)[n].open) // just return the number if already open
		return n;
	// otherwise, we need to open it
	mm = midiInOpen(&((*inlh)[n].handle), n, (DWORD)my_read_hook, NULL, CALLBACK_FUNCTION);
	if (mm != MMSYSERR_NOERROR)
	{
		ew.w( "couldn't open connection" ); return 0;
	}
	mm = midiInStart((*inlh)[n].handle);
	if (mm != MMSYSERR_NOERROR)
	{
		ew.w( "couldn't start connection" ); return 0;
	}
	(*inlh)[n].open = 1; // remember that we opened it
	return (ioref)n;
}

ioref
OMS_app::
output_refnum( const char* name )
{
	err_ind e;
	MMRESULT mm;
	UINT n = default_output_ioref;;
	e = OMS_node_info<MyMidiOut>::select_by_name( name, onlh, &n);
	if (e)
	{
		ew.w( "couldn't select an output node" ); return 0;
	}
	if ((*onlh)[n].open) // just return the number if already open
		return n;
	// otherwise, we need to open it
	mm = midiOutOpen(&((*onlh)[n].handle), n, NULL, NULL, CALLBACK_NULL);
	if (mm != MMSYSERR_NOERROR)
	{
		ew.w( "couldn't open connection" ); return 0;
	}
	(*onlh)[n].open = 1; // remember that we opened it
	return (ioref)n;
}

void
OMS_app::
closeAllInputs()
{
	for (int i = 0; i < inlh->length; i++)
		if ((*inlh)[i].open)
		{
			if ((midiInStop((*inlh)[i].handle) == MMSYSERR_NOERROR) && (midiInClose((*inlh)[i].handle) == MMSYSERR_NOERROR))
				(*inlh)[i].open = 0;
			else
				ew.w("unable to close '%s'", (*inlh)[i].szPname);
		}
}

void
OMS_app::
closeAllOutputs()
{
	for (int i = 0; i < onlh->length; i++)
		if ((*onlh)[i].open)
		{
			if (midiOutClose((*onlh)[i].handle) == MMSYSERR_NOERROR)
				(*onlh)[i].open = 0;
			else
				ew.w("unable to close '%s'", (*onlh)[i].szPname);
		}
}

int
OMS_app::
howManyOutputNodes()
{
	return onlh->length;
}

ioref
OMS_app::
getOutputRefnum(int i)
{
	return (ioref)i;
}


void CALLBACK
OMS_app::
my_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	if (wMsg == MIM_DATA)
	{
		if (lp && lp->handles(dwParam1))
		{
			UINT n = 0;
			for (int i = 0; i < oap->inlh->length; i++)
				if ((*(oap->inlh))[i].open && (*(oap->inlh))[i].handle == hMidiIn)
					n = (UINT)i;
			MyQueueElement q;
			q.dwParam1 = dwParam1;
			q.dwParam2 = dwParam2;
			q.n = n;
			cq.insertp(&q);
			schedp->processQueue(); // ask schedule thread to call our processQueue
		}
	}
}

void OMS_app::
processQueue()
{
	MyQueueElement *q;
	int i = 0;
	while ( (q = cq.headp()) != NULL )
	{
		if (lp && lp->handles(q->dwParam1)) // added this for safety 11/12/95 PP
			lp->parse( q->dwParam1, q->dwParam2, q->n );
		cq.remove_head();
		i++;
	}
	if (i > 0 && statistics_on)
		queueStat.throwInBin(i); // added for statistics gathering 11/12/95 PP
}

void
OMS_app::
write( const MIDI_msg *m )
{
		DWORD p;
		UINT n;
		m->to_OMS( &p, &n );	
		MMRESULT mm;
		mm = midiOutShortMsg((*onlh)[n].handle, p);
		if (mm != MMSYSERR_NOERROR)
		{
			char errtext[MAXERRORLENGTH+1];
			midiOutGetErrorText(mm, errtext, MAXERRORLENGTH);
			ew.w("output error -- %s",errtext);
		}
}

//--------------------------------------------------------------------------------------------------

#include "string.h" // for strstr()

template <class T>
err_ind
OMS_node_info<T>::
select_by_name( const char *name, TArrayPtr nlh, UINT *id )
{
	UINT num_nodes = nlh->length;

	char name_upper[MAXPNAMELEN+1];
	strtoupper(name, name_upper);

	int none_found = 1;
	
	for (UINT i = 0; i < num_nodes && none_found; i++)
	{
		char tmp[MAXPNAMELEN+1];
		strtoupper( (*nlh)[i].szPname, tmp);
				
		if (strstr(tmp, name_upper))
		{
			tw.pr( "SUCCESS: \"%s\" matched to \"%s\"\r", name, (*nlh)[i].szPname );
			*id = i;
			none_found = 0;
		}
	}
	if (none_found)
		if (num_nodes > 0)
		{	
			ew.w( "Couldn't find \"%s\", using \"%s\" instead", name, (*nlh)[*id].szPname );
		}
		else { ew.w("No input nodes exist"); return err; }
		
	return no_err;
}

template <class T>
void
OMS_node_info<T>::
print_nodes( TArrayPtr nlh, const char *type )
{
	UINT num_nodes = nlh->length;

	tw.pr( "\rnumber of %s nodes: %d\r\r", type, num_nodes );
	for (UINT i=0; i<num_nodes; i++ )
		tw.pr( "%d %s\r", i, (*nlh)[i].szPname );
}

// Authors: Ben Denckla, Herb Yang, Patrick Pelletier
