/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

// The constructor of the class "Filedata" interprets a MIDI file and stores the data in the
// Filedata object.  

#include "Filedata.h"
#include <stdio.h>

static long read_var_len(uchar **ch) 
{
	long value;
	uchar	h, *c = *ch;
	
	if ((value = *c++) & 0x80)
	{
		value &= 0x7f;
		do value = (value << 7) + ((h = *c++) & 0x7f); while (h & 0x80);
	}
	*ch = c;
	return value;
}
/* Returns the value of the variable length quantity at *ch, moving *ch along as well.
Code adapted from p.144 of MIDI Programmer's Handbook by DeFuria & Scacciaferro. */

static bool read_chan_msg( MidiMsg* m, uchar **ch )
{
	static uchar rs;
	uchar *c = *ch;
	
	if (*c & 0x80)
	{
		if (MidiMsg::msgt(*c) == MidiMsg::system) return false;
		m->s = rs = *c++;	// update running status
	}
	else m->s = rs;						// use running status
	m->d1 = *c++; 							// all channel msgs have at least 1 data byte
	if (m->len() == 3) m->d2 = *c++;
	*ch = c;
	return true;
}
/* If there is a channel msg at *ch, reads it into m and returns true, moving *ch along as well.
Otherwise, returns false. */

int Filedata::dual_purpose_read(uchar* in_raw, track* t, int byte_count, void (*warnfunc)(const char*))
{
	int event_num = 0;
	uchar *raw = in_raw;
	MidiMsg	m;
	
	if (t) // figure out number of MidiMsgs (using weird recursive hack)
		t->m = new MidiMsgArray(dual_purpose_read(in_raw, NULL, byte_count, warnfunc));
	while (raw - in_raw < byte_count)
	{
		m.time = read_var_len(&raw);
		if (!read_chan_msg( &m, &raw ))
		{	
			int st = MidiMsg::smsgt(*raw++);			
			long len = 0;
			uchar type;
			switch (st)
			{
			case smsgt::file_meta:
				type = *raw++;
				len = read_var_len(&raw);
				if (t) m.set_meta(type, raw, len);
				break;
			case smsgt::sys_ex:
				len = read_var_len(&raw);
				if (t) m.set_sys_ex(raw, len);
				break;
			case smsgt::file_escape:
				warnfunc( "MIDI file contains an escape: not implemented yet" );
				return event_num;
			default:
				warnfunc( "MIDI file contains an illegal system message" );
				return event_num;
			}
			raw += len;
		}
		if (t) (*(t->m))[event_num] = m;
		event_num++;
	}
	return event_num;
}

Filedata::
Filedata(const char* midi_file_name, void (*warnfunc)(const char*))
{
	FILE *ifp = fopen(midi_file_name, "rb");
	if (!ifp)
	{
		char* s = new char[80+strlen(midi_file_name)];
		hdr.ntrks = 0;
		sprintf(s, "Couldn't open MIDI file '%s'\r", midi_file_name );
		warnfunc(s);
		delete[] s;
	}
	else
	{
		fread(&hdr, sizeof(hdr), 1, ifp);
		// swap bytes because we are little-endian
		hdr.chunk_type = (int4) swap4((uint4)hdr.chunk_type);
		hdr.length = (int4) swap4((uint4)hdr.length);
		hdr.format = (int2) swap2((uint2)hdr.format);
		hdr.ntrks = (int2) swap2((uint2)hdr.ntrks);
		hdr.division = (int2) swap2((uint2)hdr.division);
		trk = new trackArray(hdr.ntrks);
		for (int i = 0; i < hdr.ntrks; i++)
		{
			fread(&((*trk)[i].hdr), sizeof(trk_hdr), 1, ifp);
			// swap bytes because we are little-endian
			(*trk)[i].hdr.chunk_type = (int4) swap4((uint4)((*trk)[i].hdr.chunk_type));
			(*trk)[i].hdr.byte_count = (int4) swap4((uint4)((*trk)[i].hdr.byte_count));			
			int n = (*trk)[i].hdr.byte_count; 
			
			uchar* raw = new uchar[n];
			fread(raw, 1, n, ifp);
			
			dual_purpose_read(raw, &((*trk)[i]), n, warnfunc);
			(*trk)[i].fixdeltas(); // change delta times to absolute times
			delete[] raw;
		}
		fclose(ifp);
	}
}

static void print_MIDImsg(MidiMsg *m, FILE *f, char* term)
{
	fprintf( f, "%s%s", (const char*)(*(m->unparse())), term );
}

void
Filedata::
printToFile(const char* file_name) const
{
	FILE *f = fopen(file_name, "w");

	fprintf
	(
		f, "\rHEADER CHUNK: file format: %d ntrks: %d division: %d\r",
		hdr.format, hdr.ntrks, hdr.division
	);
	for (int t = 0; t < hdr.ntrks; t++)
	{ 
		fprintf(f, "\rTRACK %d:\r", t+1);
		if ((*trk)[t].m->length > 0)
			for (int m = 0; m < (*trk)[t].m->length; m++) print_MIDImsg(&((*((*trk)[t].m))[m]), f, "\r");
	}
	fclose(f);
}

void
Filedata::track::
fixdeltas()
{
	long totalTime = 0;
	for (int i = 0; i < m->length; i++)
	{
		totalTime += (*m)[i].time;
		(*m)[i].time = totalTime;
	}
}

/*
Meta events and Sys Ex messages are stored with a system message status byte of type "ext"
(not part of the MIDI standard, I borrowed and unused system message type).
The data is stored in the ext_data structure pointed to by ed.

Authors: Herb Yang, Ben Denckla, Patrick Pelletier
*/